import os
from abc import ABC, abstractmethod


class FileSystemObject(os.PathLike, ABC):
    def __init__(self, path, *args, **kwargs):
        self._path = str(path).strip()

    @abstractmethod
    def _validate_path(self, path) -> bool:
        raise NotImplementedError

    def __fspath__(self):
        return self._path

    @abstractmethod
    def size(self) -> int:
        raise NotImplementedError

    def basename(self) -> str:
        return os.path.basename(self._path)

    @abstractmethod
    def exists(self) -> bool:
        raise NotImplementedError

    def is_file_path(self) -> bool:
        return False

    def is_directory_path(self) -> bool:
        return False


class File(FileSystemObject, ABC):
    @abstractmethod
    def write_text(self, content, mode, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def write_bytes(self, content, mode, encoding="utf-8", *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def touch(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def remove(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def read_text(self):
        raise NotImplementedError

    @abstractmethod
    def read_bytes(self):
        raise NotImplementedError

    @abstractmethod
    def checksum(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def duplicate(self, dst: "File"):
        raise NotImplementedError

    @abstractmethod
    def copy_to(self, dst: "Directory"):
        raise NotImplementedError


class DirectoryPrefix(FileSystemObject, ABC):
    @abstractmethod
    def remove(self, missing_ok: bool = True, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def list(self, recursive: bool = True, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def copy_to(self, dst: "Directory"):
        raise NotImplementedError

    @abstractmethod
    def zip_to(self, dst: File):
        raise NotImplementedError

    @abstractmethod
    def tar_gz_to(self, dst: File):
        raise NotImplementedError

    @abstractmethod
    def file_count(self) -> int:
        raise NotImplementedError


class Directory(DirectoryPrefix, ABC):
    @abstractmethod
    def create(self, parents: bool = True, exist_ok: bool = False, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def duplicate(self, dst: "Directory"):
        raise NotImplementedError

    @abstractmethod
    def joinpath(self, *other):
        raise NotImplementedError


class FileSystem:
    @staticmethod
    @abstractmethod
    def to_file(path):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_directory(path):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def to_directory_prefix(path):
        raise NotImplementedError


FS = FileSystem



def convert_mode(mode: str, format: str) -> str:
    mode = mode.upper()
    m = "r"
    if mode == "READ":
        m = "r"
    elif mode in ("WRITE", "OVERWRITE"):
        m = "w"
    elif mode == "APPEND":
        m = "a"
    else:
        raise RuntimeError(f"Un-supported mode = {mode}")

    if format == "bytes":
        return f"{m}b"
    elif format == "text":
        return m

    raise RuntimeError(f"Invalid format: {format}")



from pathlib import PosixPath

from ufs.base import Directory, File


class PosixDirectory(Directory):
    def is_directory_path(self) -> bool:
        return True

    def create(self, parents: bool = True, exist_ok: bool = False, *args, **kwargs):
        PosixPath(self).mkdir(parents=parents, exist_ok=exist_ok)

    def duplicate(self, dst: "Directory"):
        pass

    def joinpath(self, *other):
        return PosixPath(self).joinpath(*other)

    def remove(self, missing_ok: bool = True, *args, **kwargs):
        PosixPath(self).unlink(missing_ok)

    def list(self, recursive: bool = True, *args, **kwargs):
        pass

    def copy_to(self, dst: "Directory"):
        pass

    def zip_to(self, dst: File):
        pass

    def tar_gz_to(self, dst: File):
        pass

    def file_count(self) -> int:
        pass

    def size(self) -> int:
        pass

    def exists(self) -> bool:
        return PosixPath(self).exists()



from ufs.base import DirectoryPrefix, File
from ufs.s3.s3_common import S3Object


class S3DirectoryPrefix(S3Object, DirectoryPrefix):
    def remove(self, *args, **kwargs):
        pass

    def list(self, recursive: bool = True, *args, **kwargs):
        pass

    def copy_to(self, dst: "Directory"):
        pass

    def zip_to(self, dst: File):
        pass

    def tar_gz_to(self, dst: File):
        pass

    def file_count(self) -> int:
        pass

    def size(self) -> int:
        pass

    def exists(self) -> bool:
        pass



from pathlib import PosixPath

from ufs.base import File
from ufs.posix.posix_common import convert_mode


class PosixFile(File):
    def is_file_path(self) -> bool:
        return True

    def write_text(self, content: str, mode, *args, **kwargs):
        mode = convert_mode(mode, "text")
        with open(self, mode) as f:
            f.write(content)

    def write_bytes(self, content: bytes, mode, encoding="utf-8", *args, **kwargs):
        mode = convert_mode(mode, "bytes")
        with open(self, mode) as f:
            f.write(content)

    def touch(self, exist_ok: bool = True, *args, **kwargs):
        PosixPath(self).touch(exist_ok=exist_ok)

    def remove(self, missing_ok: bool = True, *args, **kwargs):
        PosixPath(self).unlink(missing_ok)

    def read_text(self) -> str:
        return open(self).read()

    def read_bytes(self) -> bytes:
        return open(self, "rb").read()

    def checksum(self, *args, **kwargs):
        pass

    def duplicate(self, dst: "File"):
        pass

    def copy_to(self, dst: "Directory"):
        pass

    def size(self) -> int:
        PosixPath(self).stat()

    def exists(self) -> bool:
        pass



from ufs.base import FileSystem


class S3FileSystem(FileSystem):
    @staticmethod
    def to_file(path):
        pass

    @staticmethod
    def to_directory(path):
        pass

    @staticmethod
    def to_directory_prefix(path):
        pass



from ufs.base import FileSystemObject


class S3Object(FileSystemObject):
    def __init__(self, path, s3_client=None, protocol: str = "s3://", *args, **kwargs):
        if protocol == "s3://":
            if path.startswith("s3a://"):
                path = path.replace("s3a://", "s3://")
        elif protocol == "s3a://":
            if path.startswith("s3://"):
                path = path.replace("s3://", "s3a://")

        super().__init__(path, *args, **kwargs)

        self._client = s3_client
        self._protocol = protocol

    def _validate_path(self, path) -> bool:
        return path.startswith(self._protocol)

    def size(self) -> int:
        pass

    def exists(self) -> bool:
        pass



from ufs.base import File
from ufs.s3.s3_common import S3Object


class S3File(S3Object, File):
    def is_file_path(self) -> bool:
        return True

    def write_text(self, content, mode, *args, **kwargs):
        pass

    def write_bytes(self, content, mode, encoding="utf-8", *args, **kwargs):
        pass

    def touch(self, *args, **kwargs):
        pass

    def remove(self, missing_ok: bool = True, *args, **kwargs):
        pass

    def read_text(self):
        pass

    def read_bytes(self):
        pass

    def checksum(self, *args, **kwargs):
        pass

    def duplicate(self, dst: "File"):
        pass

    def copy_to(self, dst: "Directory"):
        pass

    def size(self) -> int:
        pass

    def exists(self) -> bool:
        pass
